#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_request.hh"
#include "loggers.hh"

int sip_codec_request::encode (const LibSip__SIPTypesAndValues::Request& msg, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode: ", (const Base_Type&)msg);

  osip_message_t* sip_message = nullptr; // TODO Use smart pointer
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to initialise internal data structures");
    return -1;
  }

  if (encode_request(msg, sip_message) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to encode Request");
    ::osip_message_free(sip_message);
    sip_message = nullptr;
    return -1;
  }
  
  char* buffer = nullptr;
  size_t length = 0;
  result = ::osip_message_to_str(sip_message, &buffer, &length);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to encode data structures");
    ::osip_message_free(sip_message);
    return -1;
  }
  loggers::get_instance().log("sip_codec_request::encode: Message:%s", buffer);
  data = OCTETSTRING(length, (const unsigned char*)buffer);
  osip_free(buffer);
  ::osip_message_free(sip_message);
  
  loggers::get_instance().log_msg("<<< sip_codec_request::encode: data=", data);
  return result;
}

int sip_codec_request::decode (const OCTETSTRING& data, LibSip__SIPTypesAndValues::Request& msg, params* p_params)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::decode: data=", data);

  // Sanity checks
  if (data.lengthof() == 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Wrong parameters");
    return -1;
  }

  osip_message_t* sip_message;
  ::parser_init();
  int result = ::osip_message_init(&sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to initialise internal data structures");
    return -1;
  }
  result = ::osip_message_parse(sip_message, (const char*)static_cast<const unsigned char*>(data), data.lengthof());
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to parse SIP message");
    ::osip_message_free(sip_message);
    return -1;
  }

  // Fill RequestLine
  decode_request_line(sip_message, msg);

  // Fill Headers
  decode_headers(sip_message, msg);

  // Fill MessageBody
  decode_message_body(sip_message, msg);
  
  // Fill Payload
  decode_payload(sip_message, msg);
  
  ::osip_message_free(sip_message);
  loggers::get_instance().log_msg("sip_codec_request::decode: ", msg);
  
  return 0;
}

int sip_codec_request::encode_request(const LibSip__SIPTypesAndValues::Request& p_request, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request");
  
  encode_request_line(p_request.requestLine(), p_sip_message);
  
  encode_request_headers(p_request.msgHeader(), p_sip_message);
  
  if (p_request.messageBody().is_present()) {
    const LibSip__MessageBodyTypes::MessageBody& m = static_cast<const LibSip__MessageBodyTypes::MessageBody&>(*p_request.messageBody().get_opt_value());
    encode_request_message_body(m, p_sip_message);
  }
  
  if (p_request.payload().is_present()) {
    const LibSip__SIPTypesAndValues::Payload& p = static_cast<const LibSip__SIPTypesAndValues::Payload&>(*p_request.payload().get_opt_value());
    encode_request_payload(p, p_sip_message);
  }

  return 0;
} // End of method encode_request

int sip_codec_request::encode_request_line(const LibSip__SIPTypesAndValues::RequestLine& p_request_line, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request_line");

  // Sip method
  std::string str(LibSip__SIPTypesAndValues::Method::enum_to_str(p_request_line.method()));
  char *p = (char*)osip_malloc(str.length() - 2 + 1); // Will be freed by osip_message_free
  ::strcpy(p, str.substr(0, str.length() - 2).c_str()); // Remove _E
  ::osip_message_set_method(p_sip_message, p);

  // Sip uri
  osip_uri_t *uri;
  if (encode_sip_url(p_request_line.requestUri(), &uri) == -1) {
    loggers::get_instance().warning("sip_codec_request::encode_request_line: Faile to encode SipUrl");
    return -1;
  }
  ::osip_message_set_uri(p_sip_message, uri);

  // SIP version
  p = (char*)osip_malloc(p_request_line.sipVersion().lengthof() + 1);
  ::strcpy(p, (char*)static_cast<const char*>(p_request_line.sipVersion()));
  ::osip_message_set_version(p_sip_message, p);

  return 0;
} // End of method encode_request_line

int sip_codec_request::encode_request_headers(const LibSip__SIPTypesAndValues::MessageHeader& p_msg_header, osip_message_t* p_sip_message)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode_request_headers: ", p_msg_header);
  
  // Encode mandatory fieds
  // From
  const LibSip__SIPTypesAndValues::From from = p_msg_header.fromField();
  osip_from_t* from_header = nullptr;
  if (encode_from_header(p_msg_header.fromField(), from_header) == -1) {
    return -1;
  }
  char* hvalue;
  ::osip_from_to_str(from_header, &hvalue);
  ::osip_message_set_from(p_sip_message, hvalue);
  ::osip_from_free(from_header);
  
  // To
  const LibSip__SIPTypesAndValues::To to = p_msg_header.toField();
  osip_to_t* to_header = nullptr;
  if (encode_to_header(p_msg_header.toField(), to_header) == -1) {
    return -1;
  }
  ::osip_to_to_str(to_header, &hvalue);
  ::osip_message_set_to(p_sip_message, hvalue);
  ::osip_to_free(to_header);

  // Via
  const LibSip__SIPTypesAndValues::Via via = p_msg_header.via();
  osip_via_t* via_header = nullptr;
  if (encode_via_header(p_msg_header.via(), via_header) == -1) {
    return -1;
  }
  ::osip_via_to_str(via_header, &hvalue);
  ::osip_message_set_via(p_sip_message, hvalue);
  ::osip_via_free(via_header);
  
  return 0;
} // End of method encode_request_headers

int sip_codec_request::encode_request_message_body(const LibSip__MessageBodyTypes::MessageBody& p_message_body, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request_message_body");

  return 0;
} // End of method encode_request_message_body

int sip_codec_request::encode_request_payload(const LibSip__SIPTypesAndValues::Payload& p_payload, osip_message_t* p_sip_message)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_request_payload");

  return 0;
} // End of method encode_request_payload

int sip_codec_request::encode_sip_url(const LibSip__SIPTypesAndValues::SipUrl& p_sip_uri, osip_uri_t** p_uri)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode_sip_uri: ", p_sip_uri);

  std::string host;
  std::string port;
  osip_uri_t *uri = nullptr;
  ::osip_uri_init(&uri);
  ::osip_uri_set_scheme(uri, (char*)static_cast<const char*>(p_sip_uri.scheme()));
  const LibSip__SIPTypesAndValues::UriComponents& components = p_sip_uri.components();
  if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_sip)) {
    const LibSip__SIPTypesAndValues::SipUriComponents& s = p_sip_uri.components().sip();
    if (s.userInfo().is_present()) {
      const LibSip__SIPTypesAndValues::UserInfo& u = static_cast<const LibSip__SIPTypesAndValues::UserInfo>(s.userInfo());
      ::osip_uri_set_username(uri, (char*)static_cast<const char*>(static_cast<CHARSTRING>(u.userOrTelephoneSubscriber())));
      if (u.password().is_present()) {
        const CHARSTRING& c = static_cast<CHARSTRING>(u.password());
        ::osip_uri_set_password(uri, (char*)static_cast<const char*>(c));
      }
    }
    encode_host_port(s.hostPort(), host, port);
    if (!host.empty()) {
      ::osip_uri_set_host(uri, (char*)host.c_str());
    }
    if (!port.empty()) {
     ::osip_uri_set_port(uri, (char*)port.c_str());
    }
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_tel)) {
    const LibSip__SIPTypesAndValues::TelUriComponents& t = p_sip_uri.components().tel();
    loggers::get_instance().error("sip_codec_request::encode_request_line: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_tel");
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_urn)) {
    const LibSip__SIPTypesAndValues::UrnUriComponents& u = p_sip_uri.components().urn();
    loggers::get_instance().error("sip_codec_request::encode_request_line: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_urn");
  } else if (components.ischosen(LibSip__SIPTypesAndValues::UriComponents::ALT_other)) {
    loggers::get_instance().error("sip_codec_request::encode_request_line: Unsupported LibSip__SIPTypesAndValues::UriComponents::ALT_other");
  } // else, noting to do
  
  if (uri != nullptr) {
    ::osip_uri_clone(uri, p_uri);
  } else {
    *p_uri = nullptr;
  }
  
  return 0;
} // End of method encode_sip_url

void sip_codec_request::encode_host_port(const LibSip__SIPTypesAndValues::HostPort& p_host_port, std::string& p_host, std::string& p_port)
{
  loggers::get_instance().log(">>> sip_codec_request::encode_host_port");
  
  if (p_host_port.host().is_present()) {
    p_host.assign(static_cast<const char*>(static_cast<CHARSTRING>(p_host_port.host())));
  } else {
    p_host.clear();
  }
  if (p_host_port.portField().is_present()) {
    p_port.assign(std::to_string(static_cast<int>(static_cast<INTEGER>(p_host_port.portField()))));
  } else {
    p_port.clear();
  }
} // End of method encode_host_port

int sip_codec_request::encode_from_header(const LibSip__SIPTypesAndValues::From& p_from, osip_from_t* p_from_header) {
  loggers::get_instance().log(">>> sip_codec_request::encode_from_header");
  
  ::osip_from_init(&p_from_header);
  const LibSip__SIPTypesAndValues::Addr__Union& a = p_from.addressField();
  if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_nameAddr)) {
    const LibSip__SIPTypesAndValues::NameAddr& addr = a.nameAddr();
    osip_uri_t *uri;
    if (encode_sip_url(addr.addrSpec(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_from_header: Failed to encode SipUrl");
      ::osip_from_free(p_from_header);
      p_from_header = nullptr;
      return -1;
    }
    ::osip_from_set_url(p_from_header, uri);
    if (addr.displayName().is_present()) {
      const LibSip__SIPTypesAndValues::DisplayName& n = static_cast<const LibSip__SIPTypesAndValues::DisplayName&>(addr.displayName());
      if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_token)) {
        ::osip_from_set_displayname(p_from_header, (char*)static_cast<const char*>(static_cast<CHARSTRING>(n.token())));
      } else if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_quotedString)) {
        ::osip_from_set_displayname(p_from_header, (char*)static_cast<const char*>(static_cast<CHARSTRING>(n.quotedString())));
      } else {
        loggers::get_instance().warning("sip_codec_request::encode_from_header: Failed to encode DisplayName");
        ::osip_from_free(p_from_header);
        p_from_header = nullptr;
        return -1;
      }
    }
  } else if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_addrSpecUnion)) {
    osip_uri_t *uri;
    if (encode_sip_url(a.addrSpecUnion(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_from_header: Failed to encode SipUrl");
      ::osip_from_free(p_from_header);
      p_from_header = nullptr;
      return -1;
    }
    ::osip_from_set_url(p_from_header, uri);
  } else {
    loggers::get_instance().warning("sip_codec_request::encode_from_header: Failed to encode Addr__Union");
    ::osip_from_free(p_from_header);
    p_from_header = nullptr;
    return -1;
  }
  if (p_from.fromParams().is_present()) {
    const LibSip__Common::SemicolonParam__List& l = static_cast<const LibSip__Common::SemicolonParam__List>(p_from.fromParams());
    // TODO encode_semi_colon_params(l, );
    loggers::get_instance().warning("sip_codec_request::encode_from_header: Failed to encode SemicolonParam__List");
    ::osip_from_free(p_from_header);
    p_from_header = nullptr;
    return -1;
  }

  return 0;
} // End of method encode_from_header

int sip_codec_request::encode_to_header(const LibSip__SIPTypesAndValues::To& p_to, osip_to_t* p_to_header) {
  loggers::get_instance().log(">>> sip_codec_request::encode_to_header");
  
  ::osip_to_init(&p_to_header);
  const LibSip__SIPTypesAndValues::Addr__Union& a = p_to.addressField();
  if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_nameAddr)) {
    const LibSip__SIPTypesAndValues::NameAddr& addr = a.nameAddr();
    osip_uri_t *uri;
    if (encode_sip_url(addr.addrSpec(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_to_header: Failed to encode SipUrl");
      ::osip_to_free(p_to_header);
      p_to_header = nullptr;
      return -1;
    }
    ::osip_to_set_url(p_to_header, uri);
    //::osip_uri_free(uri);
    if (addr.displayName().is_present()) {
      const LibSip__SIPTypesAndValues::DisplayName& n = static_cast<const LibSip__SIPTypesAndValues::DisplayName&>(addr.displayName());
      if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_token)) {
        ::osip_to_set_displayname(p_to_header, (char*)static_cast<const char*>(static_cast<CHARSTRING>(n.token())));
      } else if (n.ischosen(LibSip__SIPTypesAndValues::DisplayName::ALT_quotedString)) {
        ::osip_to_set_displayname(p_to_header, (char*)static_cast<const char*>(static_cast<CHARSTRING>(n.quotedString())));
      } else {
        loggers::get_instance().warning("sip_codec_request::encode_to_header: Failed to encode DisplayName");
        ::osip_to_free(p_to_header);
        p_to_header = nullptr;
        return -1;
      }
    }
  } else if (a.ischosen(LibSip__SIPTypesAndValues::Addr__Union::ALT_addrSpecUnion)) {
    osip_uri_t *uri;
    if (encode_sip_url(a.addrSpecUnion(), &uri) == -1) {
      loggers::get_instance().warning("sip_codec_request::encode_to_header: Failed to encode SipUrl");
      ::osip_to_free(p_to_header);
      p_to_header = nullptr;
      return -1;
    }
    ::osip_to_set_url(p_to_header, uri);
  } else {
    loggers::get_instance().warning("sip_codec_request::encode_to_header: Failed to encode Addr__Union");
    ::osip_to_free(p_to_header);
    p_to_header = nullptr;
    return -1;
  }
  if (p_to.toParams().is_present()) {
    const LibSip__Common::SemicolonParam__List& l = static_cast<const LibSip__Common::SemicolonParam__List>(p_to.toParams());
    // TODO encode_semi_colon_params(l, );
    loggers::get_instance().warning("sip_codec_request::encode_to_header: Failed to encode SemicolonParam__List");
    ::osip_to_free(p_to_header);
    p_to_header = nullptr;
    return -1;
  }

  return 0;
} // End of method encode_to_header

int sip_codec_request::encode_via_header(const LibSip__SIPTypesAndValues::Via& p_via, osip_via_t* p_via_header) {
  loggers::get_instance().log(">>> sip_codec_request::encode_via_header");
  
  ::osip_via_init(&p_via_header);
  const LibSip__SIPTypesAndValues::ViaBody__List& l = p_via.viaBody();
  for (int i = 0; i < l.size_of(); i++) {
    const LibSip__SIPTypesAndValues::ViaBody v = l[i];

    ::via_set_protocol(p_via_header, (char*)static_cast<const char *>(v.sentProtocol().protocolName()));
    ::via_set_version(p_via_header, (char*)static_cast<const char *>(v.sentProtocol().protocolVersion()));
    ::via_set_comment(p_via_header, (char*)static_cast<const char *>(v.sentProtocol().transport()));
    std::string host;
    std::string port;
    encode_host_port(v.sentBy(), host, port);
    if (!host.empty()) {
      ::via_set_host(p_via_header, (char*)host.c_str());
    }
    if (!port.empty()) {
      ::via_set_port (p_via_header, (char*)port.c_str());
    }
    
    if (v.viaParams().is_present()) {
      const LibSip__Common::SemicolonParam__List& params = static_cast<LibSip__Common::SemicolonParam__List>(v.viaParams());
      // TODO encode_semi_colon_params(l, );
      loggers::get_instance().warning("sip_codec_request::encode_via_header: Failed to encode SemicolonParam__List");
      ::osip_via_free(p_via_header);
      p_via_header = nullptr;
      return -1;
    }
    
  } // End of 'for' statement

  return 0;
} // End of method encode_via_header









 
void sip_codec_request::decode_headers(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_headers");
  
  LibSip__SIPTypesAndValues::MessageHeader headers;
  // Fill mandatory fields
  // From
  LibSip__SIPTypesAndValues::From from_header;
  decode_from_header(::osip_message_get_from(p_sip_message), from_header);
  if (from_header.is_value()) {
    headers.fromField() = from_header;
  } else {
    headers.fromField().set_to_omit();
  }
  // To
  LibSip__SIPTypesAndValues::To to_header;
  decode_to_header(::osip_message_get_to(p_sip_message), to_header);
  if (to_header.is_value()) {
    headers.toField() = to_header;
  } else {
    headers.toField().set_to_omit();
  }
  // Via
  LibSip__SIPTypesAndValues::Via via_header;
  decode_via_header(p_sip_message, via_header);
  if (via_header.is_value()) {
    headers.via() = via_header;
  } else {
    headers.via().set_to_omit();
  }

  // Fill Optional fields
  LibSip__SIPTypesAndValues::Accept accept_header;
  decode_accept_header(p_sip_message, accept_header);
  if (accept_header.is_value()) {
    headers.accept() = accept_header;
  } else {
    headers.accept().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptContact accept_contact_header;
  decode_accept_contact_header(p_sip_message, accept_contact_header);
  if (accept_contact_header.is_value()) {
    headers.acceptContact() = accept_contact_header;
  } else {
    headers.acceptContact().set_to_omit();
  }
  
  LibSip__SIPTypesAndValues::AcceptEncoding accept_encoding_header;
  decode_accept_encoding_header(p_sip_message, accept_encoding_header);
  if (accept_encoding_header.is_value()) {
    headers.acceptEncoding() = accept_encoding_header;
  } else {
    headers.acceptEncoding().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptLanguage accept_language_header;
  decode_accept_language_header(p_sip_message, accept_language_header);
  if (accept_language_header.is_value()) {
    headers.acceptLanguage() = accept_language_header;
  } else {
    headers.acceptLanguage().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AlertInfo alert_info_header;
  decode_alert_info_header(p_sip_message, alert_info_header);
  if (alert_info_header.is_value()) {
    headers.alertInfo() = alert_info_header;
  } else {
    headers.alertInfo().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Allow allow_header;
  decode_allow_header(p_sip_message, allow_header);
  if (allow_header.is_value()) {
    headers.allow() = allow_header;
  } else {
    headers.allow().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AllowEvents allow_events_header;
  decode_allow_events_header(p_sip_message, allow_events_header);
  if (allow_events_header.is_value()) {
    headers.allowEvents() = allow_events_header;
  } else {
    headers.allowEvents().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Authorization authorization_header;
  decode_authorization_header(p_sip_message, authorization_header);
  if (authorization_header.is_value()) {
    headers.authorization() = authorization_header;
  } else {
    headers.authorization().set_to_omit();
  }
  
  headers.authenticationInfo().set_to_omit();

  LibSip__SIPTypesAndValues::CallId call_id_header;
  decode_call_id_header(p_sip_message, call_id_header);
  if (call_id_header.is_value()) {
    headers.callId() = call_id_header;
  } else {
    headers.callId().set_to_omit();
  }
  
  headers.callInfo().set_to_omit();
  headers.contact().set_to_omit();
  headers.contentDisposition().set_to_omit();
  headers.contentEncoding().set_to_omit();
  headers.contentLanguage().set_to_omit();
  
  LibSip__SIPTypesAndValues::ContentLength content_length_header;
  decode_content_length_header(p_sip_message, content_length_header);
  if (content_length_header.is_value()) {
    headers.contentLength() = content_length_header;
  } else {
    headers.contentLength().set_to_omit();
  }

  LibSip__SIPTypesAndValues::ContentType content_type_header;
  decode_content_type_header(p_sip_message, content_type_header);
  if (content_type_header.is_value()) {
    headers.contentType() = content_type_header;
  } else {
    headers.contentType().set_to_omit();
  }

  LibSip__SIPTypesAndValues::CSeq c_seq_header;
  decode_c_seq_header(p_sip_message, c_seq_header);
  if (c_seq_header.is_value()) {
    headers.cSeq() = c_seq_header;
  } else {
    headers.cSeq().set_to_omit();
  }
  
  headers.date().set_to_omit();
  headers.errorInfo().set_to_omit();
  headers.event().set_to_omit();
  headers.expires().set_to_omit();
  headers.featureCaps().set_to_omit();
  headers.geolocation().set_to_omit();
  headers.geolocationRouting().set_to_omit();
  headers.historyInfo().set_to_omit();
  headers.infoPackage().set_to_omit();
  headers.inReplyTo().set_to_omit();

  LibSip__SIPTypesAndValues::MaxForwards max_forwards_header;
  decode_max_forwards_header(p_sip_message, max_forwards_header);
  if (max_forwards_header.is_value()) {
    headers.maxForwards() = max_forwards_header;
  } else {
    headers.maxForwards().set_to_omit();
  }

  headers.mimeVersion().set_to_omit();
  headers.minExpires().set_to_omit();
  headers.minSE().set_to_omit();
  headers.organization().set_to_omit();
  headers.pAccessNetworkInfo().set_to_omit();
  headers.pAssertedID().set_to_omit();
  headers.pAssertedService().set_to_omit();
  headers.pAssociatedURI().set_to_omit();
  headers.path().set_to_omit();
  headers.pCalledPartyID().set_to_omit();
  headers.pChargingFunctionAddresses().set_to_omit();
  headers.pChargingVector().set_to_omit();
  headers.pEarlyMedia().set_to_omit();
  headers.pMediaAuthorization().set_to_omit();
  headers.pPreferredID().set_to_omit();
  headers.pPreferredService().set_to_omit();
  headers.priority().set_to_omit();
  headers.privacy().set_to_omit();
  headers.proxyAuthenticate().set_to_omit();
  headers.proxyAuthorization().set_to_omit();
  headers.proxyRequire().set_to_omit();
  headers.pVisitedNetworkID().set_to_omit();
  headers.rAck().set_to_omit();
  headers.rSeq().set_to_omit();
  headers.reason().set_to_omit();
  headers.recordRoute().set_to_omit();
  headers.recvInfo().set_to_omit();
  headers.requestDisposition().set_to_omit();
  headers.referredBy().set_to_omit();
  headers.referTo().set_to_omit();
  headers.referSub().set_to_omit();
  headers.replaces().set_to_omit();
  headers.replyTo().set_to_omit();
  headers.require().set_to_omit();
  headers.retryAfter().set_to_omit();
  headers.route().set_to_omit();
  headers.securityClient().set_to_omit();
  headers.securityServer().set_to_omit();
  headers.securityVerify().set_to_omit();
  headers.server().set_to_omit();
  headers.serviceRoute().set_to_omit();
  headers.sessionExpires().set_to_omit();
  headers.sessionId().set_to_omit();
  headers.sipETag().set_to_omit();
  headers.sipIfMatch().set_to_omit();
  headers.subject().set_to_omit();
  headers.subscriptionState().set_to_omit();
  
  LibSip__SIPTypesAndValues::Supported supported_header;
  decode_supported_header(p_sip_message, supported_header);
  if (supported_header.is_value()) {
    headers.supported() = supported_header;
  } else {
    headers.supported().set_to_omit();
  }
  
  headers.timestamp().set_to_omit();
  headers.unsupported().set_to_omit();
  headers.userToUser().set_to_omit();
  headers.userAgent().set_to_omit();
  headers.warning().set_to_omit();
  headers.wwwAuthenticate().set_to_omit();
  headers.resourcePriority().set_to_omit();
  headers.answerMode().set_to_omit();
  headers.privAnswerMode().set_to_omit();
  headers.targetDialog().set_to_omit();
  headers.pAnswerState().set_to_omit();
  headers.undefinedHeader__List().set_to_omit();
  
  p_request.msgHeader() = headers;
  
  loggers::get_instance().log_msg("sip_codec_request::decode_headers: ", p_request);
} // End of method decode_headers

void sip_codec_request::decode_message_body(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_message_body");

  p_request.messageBody().set_to_omit();
} // End of method decode_message_body

void sip_codec_request::decode_payload(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_payload");

  p_request.payload().set_to_omit();
  
} // End of method decode_payload

void sip_codec_request::decode_semi_colon_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::SemicolonParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_request::decode_semi_colon_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::SemicolonParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_request::decode_semi_colon_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_semi_colon_params: ", p_list);
} // End of method decode_semi_colon_params

void sip_codec_request::decode_ampersand_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::AmpersandParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_request::decode_ampersand_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::AmpersandParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_request::decode_ampersand_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::AmpersandParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_ampersand_params: ", p_list);
} // End of method decode_ampersand_params

void sip_codec_request::decode_option_tag_list(const char* p_list, OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>& p_options_tags)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_option_tag_list");
  
  if (p_list != nullptr) {
    LibSip__SIPTypesAndValues::OptionTag__List l;
    std::string str(p_list);
    std::size_t i = str.find(","); // See RFC 3261 - Page 231 / LibSip__SIPTypesAndValues::OptionTag__List
    if (i == std::string::npos) { // Only on item
      l[0] = CHARSTRING(p_list);
    } else {
      int idx = 0;
      while(i != std::string::npos) {
        l[idx++] = CHARSTRING(str.substr(0, i - 1).c_str());
        str = str.substr(i);
        i = str.find(",");
      } // End of 'while' statement
    }
    p_options_tags = OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>(l);
  } else {
    p_options_tags.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_option_tag_list: ", p_options_tags);
} // End of method decode_option_tag_list

void sip_codec_request::decode_host_port(const char* p_host, const char* p_port, LibSip__SIPTypesAndValues::HostPort& p_host_port) {
  loggers::get_instance().log(">>> sip_codec_request::decode_host_port");
  
  if (p_host == nullptr) {
    p_host_port.host().set_to_omit();
  } else {
    p_host_port.host() = OPTIONAL<CHARSTRING>(CHARSTRING(p_host));
  }
  OPTIONAL<INTEGER> port;
  if (p_port == nullptr) {
    p_host_port.portField().set_to_omit();
  } else {
    p_host_port.portField() = OPTIONAL<INTEGER>(INTEGER(std::stoi(p_port)));
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_host_port: ", p_host_port);
} // End of method decode_host_port"

void sip_codec_request::decode_request_line(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_request_line");
  
  LibSip__SIPTypesAndValues::RequestLine request_line;
  std::string str(::osip_message_get_method(p_sip_message));
  str += "_E";
  request_line.method() = LibSip__SIPTypesAndValues::Method(LibSip__SIPTypesAndValues::Method::str_to_enum(str.c_str()));
  loggers::get_instance().log_msg("sip_codec_request::decode_request_line: Method: ", request_line.method());
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_message_get_uri(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_request::decode_request_line: SipUrl: ", uri);
  request_line.requestUri() = uri;
  request_line.sipVersion() = CHARSTRING(::osip_message_get_version(p_sip_message));
  p_request.requestLine() = request_line;
  // FIXME To be continued

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_request_line: ", p_request);
}

void sip_codec_request::decode_uri(LibSip__SIPTypesAndValues::SipUrl& p_sip_url, const osip_uri_t* p_uri) {
  loggers::get_instance().log(">>> sip_codec_request::decode_uri");
  
  // Scheme
  LibSip__SIPTypesAndValues::UriComponents uri_components;
  if (::osip_uri_get_scheme((osip_uri_t*)p_uri) != nullptr) {
    p_sip_url.scheme() = CHARSTRING(::osip_uri_get_scheme((osip_uri_t*)p_uri));
    loggers::get_instance().log_msg("sip_codec_request::decode_uri: scheme", p_sip_url.scheme());
    // SipUriComponents
    if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "sip") == 0) {
      // User Info
      OPTIONAL<LibSip__SIPTypesAndValues::UserInfo> user_info;
      if (::osip_uri_get_username((osip_uri_t*)p_uri) == nullptr) {
        user_info.set_to_omit();
      } else {
        OPTIONAL<CHARSTRING> password;
        if (::osip_uri_get_password((osip_uri_t*)p_uri) == nullptr) {
          password.set_to_omit();
        } else {
          password = CHARSTRING(::osip_uri_get_password((osip_uri_t*)p_uri));
        }
        LibSip__SIPTypesAndValues::UserInfo u(CHARSTRING(::osip_uri_get_username((osip_uri_t*)p_uri)), password);
        user_info = OPTIONAL<LibSip__SIPTypesAndValues::UserInfo>(u);
      }
      loggers::get_instance().log_msg("sip_codec_request::decode_uri: user_info: ", user_info);
      // HostPort
      LibSip__SIPTypesAndValues::HostPort host_port;
      decode_host_port(::osip_uri_get_host((osip_uri_t*)p_uri), ::osip_uri_get_port((osip_uri_t*)p_uri), host_port);
      loggers::get_instance().log_msg("sip_codec_request::decode_uri: host_port: ", host_port);
      uri_components.sip() = LibSip__SIPTypesAndValues::SipUriComponents(user_info, host_port);
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "tel") == 0) {
      // FIXME uri_components.tel() = LibSip__SIPTypesAndValues::TelUriComponents(user_info, host_port);
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "urn") == 0) {
      // FIXME uri_components.urn() = LibSip__SIPTypesAndValues::UrnUriComponents(user_info, host_port);
    } else {
      char *buffer = nullptr;
      ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
      uri_components.other() = CHARSTRING(buffer);
      osip_free(buffer); // Macro
    }
  } else {
    char *buffer = nullptr;
    ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
    uri_components.other() = CHARSTRING(buffer);
    osip_free(buffer); // Macro
  }
  p_sip_url.components() = uri_components;
  // UrlParameters
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_uri->url_params, params);
  p_sip_url.urlParameters() = params;
  // Headers
  OPTIONAL<LibSip__Common::AmpersandParam__List> headers;
  decode_ampersand_params(p_uri->url_headers, headers);
  p_sip_url.headers() = headers;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_uri: ", p_sip_url);
} // End of method decode_uri

void sip_codec_request::decode_accept_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Accept& p_accept_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_accept_header: %p", p_sip_message->accepts);

  // Sanity checks
  osip_accept_t *sip_accept = nullptr;
  ::osip_message_get_accept(p_sip_message, 0, &sip_accept);
  if (sip_accept == nullptr) {
    return;
  }
  
  // FieldName
  p_accept_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_E"));
  // AcceptArgs
  // TODO To be done
  p_accept_header.acceptArgs().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_accept_header: ", p_accept_header);
} // End of method decode_accept_header

void sip_codec_request::decode_accept_contact_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptContact& p_accept_contact_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_accept_contact_header: %p", p_sip_message->contacts);

  // Sanity checks
  osip_header_t *sip_accept_contact = nullptr;
  osip_message_header_get_byname(p_sip_message,(const char *)"accept-contact", 0, &sip_accept_contact);
  if (sip_accept_contact == nullptr) {
    return;
  }
  
  // FieldName
  p_accept_contact_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_CONTACT_E"));
  // AcRcValue
  LibSip__SIPTypesAndValues::AcRcValue__List l;
  // TODO To be done
  l[0] = LibSip__SIPTypesAndValues::AcRcValue(CHARSTRING("*"), OPTIONAL<LibSip__Common::SemicolonParam__List>(OMIT_VALUE));
  p_accept_contact_header.acValues() = l;

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_accept_contact_header: ", p_accept_contact_header);
} // End of method decode_accept_contact_header

void sip_codec_request::decode_accept_encoding_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptEncoding& p_accept_encoding_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_accept_encoding_header: %p", p_sip_message->accept_encodings);

  // Sanity checks
  osip_accept_encoding_t *sip_accept_encoding = nullptr;
  ::osip_message_get_accept_encoding(p_sip_message, 0, &sip_accept_encoding);
  if (sip_accept_encoding == nullptr) {
    return;
  }
  // FieldName
  p_accept_encoding_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_ENCODING_E"));
  // ContentCoding
  p_accept_encoding_header.contentCoding().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_accept_encoding_header: ", p_accept_encoding_header);
} // End of method decode_accept_encoding_header

void sip_codec_request::decode_accept_language_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptLanguage& p_accept_language_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_accept_language_header: %p", p_sip_message->accept_languages);

  // Sanity checks
  osip_accept_language_t *sip_accept_language = nullptr;
  ::osip_message_get_accept_language(p_sip_message, 0, &sip_accept_language);
  if (sip_accept_language == nullptr) {
    return;
  }
  
  // FieldName
  p_accept_language_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_LANGUAGE_E"));
  // LanguageBody
  p_accept_language_header.languageBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_accept_language_header: ", p_accept_language_header);
} // End of method decode_accept_language_header

void sip_codec_request::decode_alert_info_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AlertInfo& p_alert_info_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_alert_info_header: %p", p_sip_message->alert_infos);

  // Sanity checks
  osip_alert_info_t *sip_alert_info = nullptr;
  ::osip_message_get_alert_info(p_sip_message, 0, &sip_alert_info);
  if (sip_alert_info == nullptr) {
    return;
  }
  
  // FieldName
  p_alert_info_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALERT_INFO_E"));
  // AlertInfoBody
  p_alert_info_header.alertInfoBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_alert_info_header: ", p_alert_info_header);
} // End of method decode_alert_info_header

void sip_codec_request::decode_allow_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Allow& p_allow_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_allow_header: %p", p_sip_message->allows);

  // Sanity checks
  osip_allow_t *sip_allow = nullptr;
  ::osip_message_get_allow(p_sip_message, 0, &sip_allow);
  if (sip_allow == nullptr) {
    return;
  }
  
  // FieldName
  p_allow_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_E"));
  // Methods
  p_allow_header.methods().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_allow_header: ", p_allow_header);
} // End of method decode_allow_header

void sip_codec_request::decode_allow_events_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AllowEvents& p_allow_events_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_allow_events_header: %p", p_sip_message->headers);

  // Sanity checks
  osip_header_t *sip_allow_events = nullptr;
  osip_message_header_get_byname(p_sip_message,(const char *)"allow_events", 0, &sip_allow_events);
  if (sip_allow_events == nullptr) {
    return;
  }
  
  // FieldName
  p_allow_events_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_EVENTS_E"));
  // EventTypes
  LibSip__SIPTypesAndValues::EventType__List l;
  // TODO To be done
  l[0] = CHARSTRING("*");
  p_allow_events_header.eventTypes() = l;

  loggers::get_instance().log_msg("<<< sip_codec_request::decode_allow_events_header: ", p_allow_events_header);
} // End of method decode_allow_events_header

void sip_codec_request::decode_authorization_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Authorization& p_authorization_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_authorization_header");

  // Sanity checks
  osip_authorization_t *sip_authorization = nullptr;
  ::osip_message_get_authorization(p_sip_message, 0, &sip_authorization);
  if (sip_authorization == nullptr) {
    return;
  }
  
  // FieldName
  p_authorization_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("AUTHORIZATION_E"));
  // CredentialsList
  LibSip__SIPTypesAndValues::CredentialsList l;
  int i = 0;
  do {
    loggers::get_instance().log("sip_codec_request::decode_authorization_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::Credentials v;
    LibSip__Common::CommaParam__List cl;
    int j = 0;
    if (::osip_authorization_get_digest(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_digest(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("Digest"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_realm(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_realm(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("Realm"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    // FIXME To be contiuened
    v.digestResponse() = cl;
    l[i++] = v;
  } while(::osip_message_get_authorization(p_sip_message, i, &sip_authorization) == 0); // End of 'do-while' statement
  p_authorization_header.body() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_authorization_header: ", p_authorization_header);
} // End of method decode_authorization_header

void sip_codec_request::decode_call_id_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::CallId& p_call_id_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_call_id_header: %p", p_sip_message->call_id);
  
  // Sanity check
  if (p_sip_message->call_id == nullptr) {
    return;
  }

  // FieldName
  p_call_id_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CALL_ID_E"));
  // CallId
  char *buffer;
  ::osip_call_id_to_str(p_sip_message->call_id, &buffer);
  if (buffer != nullptr) {
    p_call_id_header.callid() = CHARSTRING(buffer);
    osip_free(buffer);
  }
} // End of method decode_call_id_header

void sip_codec_request::decode_c_seq_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::CSeq& p_c_seq_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_c_seq_header: %p", p_sip_message->cseq);
  
  // Sanity check
  if (p_sip_message->cseq == nullptr) {
    return;
  }

  // FieldName
  p_c_seq_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("C_SEQ_E"));
  // Sequence number
  p_c_seq_header.seqNumber() = INTEGER(std::stoi(osip_cseq_get_number(p_sip_message->cseq)));
  // Method
  p_c_seq_header.method() = CHARSTRING(::osip_cseq_get_method(p_sip_message->cseq));
} // End of method decode_c_seq_header

void sip_codec_request::decode_content_length_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::ContentLength& p_content_length_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_content_length_header: %p", p_sip_message->content_length);
  
  // Sanity check
  if (p_sip_message->content_length == nullptr) {
    return;
  }

  // FieldName
  p_content_length_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CONTENT_LENGTH_E"));
  // Length
  char *buffer;
  ::osip_content_length_to_str(p_sip_message->content_length, &buffer);
  if (buffer != nullptr) {
    p_content_length_header.len() = INTEGER(std::stoi(buffer));
    osip_free(buffer);
  }
} // End of method decode_content_length_header

void sip_codec_request::decode_content_type_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::ContentType& p_content_type_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_content_type_header: %p", p_sip_message->content_type);
  
  // Sanity check
  if (p_sip_message->content_type == nullptr) {
    return;
  }

  // FieldName
  p_content_type_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("CONTENT_TYPE_E"));
  // SubType
  if (p_sip_message->content_type->subtype != nullptr) {
    p_content_type_header.mTypeSubtype() = CHARSTRING(p_sip_message->content_type->subtype);
  }
  // Parameters
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_message->content_type->gen_params, params);
  p_content_type_header.mParams() = params;
} // End of method decode_content_type_header


















void sip_codec_request::decode_from_header(const osip_from_t* p_sip_from, LibSip__SIPTypesAndValues::From& p_from_header)
{
  loggers::get_instance().log("sip_codec_request::decode_from_header");
  
  // Sanity check
  if (p_sip_from == nullptr) {
    return;
  }

  // FieldName
  p_from_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("FROM_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_from_get_url((osip_from_t*)p_sip_from));
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_from_get_displayname((osip_from_t*)p_sip_from) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_from_get_displayname((osip_from_t*)p_sip_from));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
  } else {
    display_name.set_to_omit();
  }
  LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
  LibSip__SIPTypesAndValues::Addr__Union addr;
  addr.nameAddr() = name_addr;
  p_from_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_from->gen_params, params);
  p_from_header.fromParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_from_header: ", p_from_header);
} // End of method decode_from_header

void sip_codec_request::decode_max_forwards_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::MaxForwards& p_max_forwards_header)
{
  loggers::get_instance().log("sip_codec_request::decode_max_forwards_header: %p", p_sip_message->headers);

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_get_max_forwards(p_sip_message, 0, &dest);
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_max_forwards_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("MAX_FORWARDS_E"));
  // Forwards
  loggers::get_instance().log("sip_codec_request::decode_max_forwards_header: got it: %s:%s", dest->hname, dest->hvalue);
  p_max_forwards_header.forwards() = INTEGER(std::stoi(dest->hvalue));
} // End of method decode_max_forwards_header

void sip_codec_request::decode_supported_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Supported& p_supported_header)
{
  loggers::get_instance().log("sip_codec_request::decode_supported_header: %p", p_sip_message->headers);

  // Sanity checks
  osip_header_t *dest = nullptr;
  osip_message_get_supported(p_sip_message, 0, &dest);
  if (dest == nullptr) {
    return;
  }
  
  // FieldName
  p_supported_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("SUPPORTED_E"));
  // OptionTags
  loggers::get_instance().log("sip_codec_request::decode_supported_header: got it: %s:%s", dest->hname, dest->hvalue);
  OPTIONAL<LibSip__SIPTypesAndValues::OptionTag__List>& l = p_supported_header.optionsTags();
  decode_option_tag_list(dest->hvalue, l);
} // End of method decode_supported_header

void sip_codec_request::decode_to_header(const osip_to_t* p_sip_to, LibSip__SIPTypesAndValues::To& p_to_header)
{
  loggers::get_instance().log("sip_codec_request::decode_to_header");
  
  // Sanity check
  if (p_sip_to == nullptr) {
    return;
  }

  // FieldName
  p_to_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("TO_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  decode_uri(uri, ::osip_to_get_url((osip_to_t*)p_sip_to));
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_to_get_displayname((osip_to_t*)p_sip_to) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_to_get_displayname((osip_to_t*)p_sip_to));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
  } else {
    display_name.set_to_omit();
  }
  LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
  LibSip__SIPTypesAndValues::Addr__Union addr;
  addr.nameAddr() = name_addr;
  p_to_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  decode_semi_colon_params(p_sip_to->gen_params, params);
  p_to_header.toParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_to_header: ", p_to_header);
} // End of method decode_to_header

void sip_codec_request::decode_via_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Via& p_via_header)
{
  loggers::get_instance().log(">>> sip_codec_request::decode_via_header: %p", p_sip_message->vias);

  // Sanity checks
  osip_via_t *sip_via = nullptr;
  ::osip_message_get_via(p_sip_message, 0, &sip_via);
  if (sip_via == nullptr) {
    return;
  }
  
  // FieldName
  p_via_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("VIA_E"));
  // ViaBody
  LibSip__SIPTypesAndValues::ViaBody__List l;
  int i = 0;
  do {
    loggers::get_instance().log("sip_codec_request::decode_via_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::ViaBody v;
    v.sentProtocol() = LibSip__SIPTypesAndValues::SentProtocol(
                                                               CHARSTRING(::via_get_protocol(sip_via)),
                                                               CHARSTRING(::via_get_version(sip_via)),
                                                               CHARSTRING(::via_get_comment(sip_via))
                                                               );
    OPTIONAL<LibSip__SIPTypesAndValues::HostPort> host_port;
    decode_host_port(::via_get_host(sip_via), ::via_get_port(sip_via), host_port);
    loggers::get_instance().log_msg("sip_codec_request::decode_via_uri: host_port: ", host_port);
    v.sentBy() = host_port;
    OPTIONAL<LibSip__Common::SemicolonParam__List> params;
    decode_semi_colon_params(sip_via->via_params, params);
    v.viaParams() = params;
    l[i++] = v;
  } while(::osip_message_get_via(p_sip_message, i, &sip_via) == 0); // End of 'while' statement
  p_via_header.viaBody() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::decode_via_header: ", p_via_header);
} // End of method decode_via_header
